
#####################################
##  FCSgate Class
#####################################
##  Methods:
##  1.  createGate
##

## Extends an FCS dataset to include a gate matrix.  This matrix
## should be describe a sequence of gating steps, so that points
## removed are not returned (that would require a "restart" of the
## gating operation).  If gate[i,j]=0, gate[i,j+k]=0, for all k>=0.


## Extends an FCS dataset to include a gate matrix

setClass("FCSgate",
         representation(gate="matrix",
                        history="vector",
                        extractGatedData.msg="vector",
                        current.data.obs="vector"),
         ## history = a vector of character strings describing
         ## what goes on from one column to the next of the gating matrix
         ## current.data.obs=positions of the
         ##    index of the current row positions of the data

         ## extractGatedData.msg: tells what actions/extractions have been done on the data
         contains=c("FCS"))

###################################################################
# Gating and related functions:
#   createGate : creating a gating index
#   icreateGate : interactive form of createGate with plots

#   gate.IPC : creating gating index for image
#              parallel coordinates plot (still need to test it)
#   extractGateHistory : obtaining the gate information
#               from the 'history' string of a FCSgate class object
#   extractGatedData : subsets/row-reduces the data matrix to
#                reflect what is selected in the gating index
#   showgate.FCS : displaying the particular gate or
#                  subset of cells that are selected
###################################################################

## CANNOT BE AUTOMATED: prev.IndexValue.In should be either 0 or 1
## depending on what value the data was subset
## from the previous index using 'extractGatedData'
## default is taking prev.IndexValue.In to be 1

## createGate will always have IndexValue.In to be 1

"createGate" <- function (x,
                          varpos=NULL,
                          gatingrange=NULL,
                          type=c("uniscut", "bidcut", "biscut", "bipcut"),
                          biscut.quadrant=c("+/+", "-/-", "-/+", "+/-"),
                          prev.gateNum=NULL, prev.IndexValue.In=NULL,
                          comment="", MY.DEBUG=FALSE) {
  
  x.name <- x[["objectname"]]
  if (is.null(varpos) || is.null(gatingrange)){
    stop("Input 'varpos' and 'gatingrange' should be numeric vectors")
  }
  
  if (!checkvars(x, MY.DEBUG=MY.DEBUG)){
    
    warning("Error with 'x' input and will attempt to fix with fixvars")
    
    x <- fixvars(x, MY.DEBUG=MY.DEBUG)
  }
  ##GATING RANGE IS ROUNDED TO integer value
  ## we are working with integer data
  if (is(x, "FCSgate")){
    if (is.null(x@current.data.obs)){
      if (length(as.vector(x@data))==0){
        stop("There is no data to createGate.")
      } else if (length(as.vector(x@data))==1){
        if (is.na(x@data)){
          stop("There is no data to createGate.")
        } else {
          stop("x@current.data.obs is null")
          
        }
      } else {
        stop("x@current.data.obs is null")
      }
    }
    if (length(x@current.data.obs) != (dim.FCS(x)[1])){
      stop(paste("Mismatch length of x@current.data.obs=",
                 x@current.data.obs,
                 "with dim.FCS(x)[1]=",
                 dim.FCS(x)[1]))
    }
  }
  
  gatingrange <- round(gatingrange, 0)
  
  ## the previous gate will be the previous column in the gating matrix,
  ## usually, but not necessarily
  
###  NOTE: 
### If any name is NA, then the name is NA or there is no name
### If any column position is NA, then there is no column position
### A column position will never have the value of "None"
  
  current.data.obs <- 1:(dim.FCS(x)[1])
  prev.IndexValue.In <- NULL
  if (is(x, "FCSgate")){
    if (!is.null(prev.gateNum)){
      if (length(as.vector(x@gate)) > 0){
        n.col.prev.gate <- prev.gateNum 
        name.subset <- colnames(x@gate)[prev.gateNum]
        prev.gate <- x@gate[,prev.gateNum]
        current.data.obs <- x@current.data.obs
        if (x@extractGatedData.msg[prev.gateNum] != "NONE"){
          ## there has been a previous subset
          ## get information about the prev.IndexValue.In
          prev.IndexValue.In <- as.numeric(unlist(strsplit(unlist(strsplit(x@extractGatedData.msg[prev.gateNum],
                                                                           split="///"))[2], split="="))[2])
        } 
      } else if (length(as.vector(x@gate))==0){
        n.col.prev.gate <- NA
        name.subset <- NA
        prev.gate <- NULL
        
      }
    } else {
      n.col.prev.gate <- NA
      name.subset <- NA
      prev.gate <- NULL
    }
  } else {
    n.col.prev.gate <- NA
    name.subset <- NA
    prev.gate <- NULL
  }
  
  
  ##get the needed data
  if (!(length(varpos) %in% c(1,2))){
    stop("Gating is implemented on one or two-dimensions only;
input 'varpos' should be of length 1 or 2.")
  }
  data <- as(x[,varpos], "matrix")
  n.row <- dim(data)[1]
  n.col <- dim(data)[2]
  
  type <- match.arg(type)
  if (n.col==1 & type !="uniscut"){
    warning(paste("Input 'varpos' is of length 1 but type =",
                  type, "which will be forced to be 'uniscut'", sep=" "))
    type <- "uniscut"
  }
  names.orig <- colnames(as(x@data, "matrix"))
  names.var <- names.orig[varpos]
  
  
  if (type=="uniscut") {
    
    Index <- ifelse(data >= gatingrange, 1, 0)
    type.descript <- paste(paste("Current.gate=uniscut(columnpos,colname)=",
                                 "(", varpos, ",", names.var, ")", sep=""), 
                           paste("COMMENT=", comment, sep=""),
                           paste("Previous.gate{gateNum,gateName}={",
                                 n.col.prev.gate, ",",  name.subset, "}", sep=""),
                           sep="///")
    gate.string <- paste( "x>=", gatingrange,sep="")
    biscut.quadrant <- ""
  } else if (type=="biscut") {
    biscut.quadrant <- match.arg(biscut.quadrant)
    
    if (biscut.quadrant=="+/+"){
      Index <- ifelse((gatingrange[1] <= data[,1]) & (gatingrange[2] <= data[,2]), 1,0)
    } else if (biscut.quadrant=="+/-"){
      Index <- ifelse((gatingrange[1] <= data[,1]) & (gatingrange[2] >= data[,2]), 1,0 )
    } else if (biscut.quadrant=="-/-"){
      Index <- ifelse((gatingrange[1] >= data[,1]) & (gatingrange[2] >= data[,2]), 1,0 )
    } else if (biscut.quadrant=="-/+"){
      Index <- ifelse((gatingrange[1] >= data[,1]) & (gatingrange[2] <= data[,2]), 1,0 )
    }
    
    type.descript <- paste(paste("Current.gate=biscut(columnpos,colnames,quadrant)=(",
                                 "c(", varpos[1], ",", varpos[2], "),",
                                 "c(", names.var[1],",", names.var[2], "),",
                                 biscut.quadrant,")", sep=""),  
                           paste("COMMENT=", comment, sep=""),
                           paste("Previous.gate{gateNum,gateName}={",
                                 n.col.prev.gate, ",", name.subset, "}", sep=""),
                           sep="///")
    gate.string <- paste("x=", gatingrange[1], ";", "y=",
                         gatingrange[2], sep="")   
    
  } else if (type=="bidcut") {
    
    Index <- ifelse((gatingrange[1] <= data[,1]) & (data[,1] <= gatingrange[2]) & (gatingrange[3] <= data[,2]) & (data[,2] <= gatingrange[4]), 1, 0) 
    biscut.quadrant <- ""
    type.descript <- paste(paste("Current.gate=bidcut(columnpos,colnames)=(",
                                 "c(", varpos[1], ",", varpos[2], "),",
                                 "c(", names.var[1],",", names.var[2], "))",sep=""), 
                           paste("COMMENT=", comment, sep=""),
                           paste("Previous.gate{gateNum,gateName}={",
                                 n.col.prev.gate, ",", name.subset, "}", sep=""),
                           sep="///")
    
    gate.string <- paste(paste("x>=", gatingrange[1], ",x<=",  gatingrange[2],sep=""),
                         paste("y>=", gatingrange[3], ",y<=", gatingrange[4], sep=""),sep=";")
    
    
    
  } else if (type=="bipcut") {
    ## convert to a matrix
    len.gtrng <- length(gatingrange)
    if (len.gtrng %% 2 !=0){
      stop("For bipcut: input 'gatingrange' should be c(c(x1,x2...xk), c(y1, y2...yk)) where length is 2k")
    }
    gatingrange <- matrix(gatingrange, ncol=2, byrow=FALSE)
    ## lib.there <- require(prada)
    ## if (lib.there==FALSE){
    ##  stop("Library prada can not be found but is needed")
    ## }
    
  
    data <- cbind(as.numeric(data[,1]), as.numeric(data[,2]))
   
    Index <-as.numeric(.Call("inPolygon", data, gatingrange ,PACKAGE="prada")) 
    
   
    type.descript <- paste(paste("Current.gate=bipcut(columnpos,colnames)=(",
                                 "c(", varpos[1], ",", varpos[2], "),",
                                 "c(", names.var[1],",", names.var[2], "))",sep=""), 
                           paste("COMMENT=", comment, sep=""),
                           paste("Previous.gate{gateNum,gateName}={",
                                 n.col.prev.gate,",", name.subset, "}", sep=""),
                           sep="///")
    
    
    gate.string <- paste("polygonal.range.x=",
                         paste(gatingrange[,1], collapse=","),
                         ";polygonal.range.y=",
                         paste(gatingrange[,2], collapse=","), sep="")
    biscut.quadrant <- ""
  }
  
  shortname.input <- paste(type, "/:/",paste(names.var, collapse="/&/"),sep="")
  type.descript <- paste(type.descript, gate.string, "IndexValue.In=1", sep="///")
  
  ## SOMETHING IS WRONG HERE 3.13.04
  ## the gate is supposed to remain the same??? YES
  ## or do we want to subset it with the data? NO
  
  ## have a slot that has the current positions in the data for the index,
  ## if we extract and then create again
  
  if (!is.null(prev.gate)){
    if (is.null(prev.IndexValue.In)){
      prev.hist <- unlist(strsplit(x@history[prev.gateNum], split="///"))
      prev.IndexValue.In <- as.numeric(unlist(strsplit(prev.hist[length(prev.hist)], split="="))[2])
    } else if (! (prev.IndexValue.In %in% min(Index):max(Index))){
      stop(paste("Input 'prev.IndexValue' is not a valid value such as <",
                 paste(unique(Index), collapse=",", sep=""), ">", sep=""))
    }

    ## only check this if already implemented 'extractGatedData' function
   ## if (length(which(prev.gate==prev.IndexValue.In))!=length(Index)) {
   ##   stop("Previous gating index is not the same length as current number of rows in data.")
  ##  }
    
    if (length(
               prev.gate)==length(Index)){
      ## make sure what is zero in the previous gate is zero in index
      Index <- ifelse(prev.gate!=prev.IndexValue.In, 0, Index)
    } else {
      ## make the lengths the same,
      ## fill with zeros, if this is a subset
      Index.pos <- which(prev.gate==prev.IndexValue.In)
      short.Index <- Index
      Index <- rep(0, length(prev.gate))
      Index[Index.pos] <- short.Index
    }
    
  }
  ## give a name to the Index
  Index <- matrix(Index, ncol=1)
  
  colnames(Index) <- paste(type, ".", paste("v", varpos, collapse="", sep=""), sep="")
  if (MY.DEBUG){
    
    cat("createGate: FCS object=", x[["objectname"]], "\n")
    cat("            type of gate (type)=", type, "\n")
    if (type=="biscut"){
      cat("            (biscut.quadrant)=", biscut.quadrant, "\n")
    }
    cat("            column position (varpos)=", varpos, "\n")
    cat("            (gatingrange)=", gate.string, "\n")
    cat("            Previous gate column position in x@gate (prev.gateNum)=", prev.gateNum, "\n")
    cat("            (comment)=", comment, "\n\n")
  }
  if (is(x, "FCSgate")){ ## tag on the new information
    if (length(Index) < (dim(x@gate)[1])){
      ## fill in with zeros

      Index.pos <- x@current.data.obs
      short.Index <- Index
      Index <- rep(0, (dim(x@gate)[1]))
      Index[Index.pos] <- short.Index
    } 
      
    x@gate <- cbind(x@gate, Index)
    
    n.col.gate.final <- dim(x@gate)[2]
    names(x@gate)[n.col.gate.final] <- shortname.input 
    x@history<- c(x@history, type.descript)
              x@extractGatedData.msg <- c(x@extractGatedData.msg, "NONE")
    x
  } else if (is(x, "FCS")){
    new("FCSgate", gate=matrix(Index, ncol=1), history=type.descript, extractGatedData.msg="NONE",
        current.data.obs=current.data.obs, x)
  }
}
            
          



## was it that the metadata remains the same?
## a new FCS object is returned
## new FCS object has reduced dimensions
## new information is added to the metadata?
## is the gate returned?  yes

##if (!isGeneric("extractGatedData")) {
##  setGeneric("extractGatedData",
##             function(x,gateNum=NULL, 
##                      IndexValue.In=1, MY.DEBUG=FALSE){

##               standardGeneric("extractGatedData")
##             })
             
##}

"extractGatedData" <- function (x, gateNum=NULL, IndexValue.In=1, MY.DEBUG=FALSE) {

  ## there are default gatenames....will only update info in the metadata if
  ## the variables have the same name
  
  ## NOTE:
  ## Will only subset the rows NOT the columns,
  ## to subset the columns of an FCS object, use [, [<-
  
  ##-------------------------
  ## input
  ##-------------------------
  ## x an object of FCSgate after createGate
  
  ## IndexValue.In  the value of the Index
  ##   (either numeric or character) which designtes inclusion in the subset
  
  ## gateNum: The column number of the gate matrix in which to subset
  
  if (length(as.vector(x@data))==0){
    stop("There is no data; cannot implement 'extractGatedData on no data")
  } else if (length(as.vector(x@data))==1){
    if (is.na(x@data)){
      stop("There is no data; cannot implement 'extractGatedData on no data")
    }
  }
  
  
  if (!checkvars(x, MY.DEBUG=MY.DEBUG)){
    warning("Error with 'x' input and will attempt to fix with fixvars")
    x <- fixvars(x, MY.DEBUG=MY.DEBUG)
  }
  
  if (length(as.vector(x@gate))==0){
    stop("There is no x@gate matrix in this input FCSgate object.")
  } else if (length(as.vector(x@gate))==1) {
    
    if (is.na(x@gate)){
      stop("There is no x@gate matrix in this input FCSgate object.")
    }
  }
  
  if (is.null(gateNum)){
    ## will take the last column by default
    gateNum <- dim(x@gate)[2]
    warning("No input 'gateNum' was specified; the last column of x@gate is chosen as the gating index.")
  }
  ## get the gating information
  ## print out as debugging statements:
  if (MY.DEBUG){
    x.name <- x[["objectname"]]
    hist.str <- unlist(x@history[gateNum])
    ##  type.descript <- unlist(strsplit(hist.str, split="/:/"))[1]
    ## names.line <- unlist(strsplit(type.descript, split="="))
    ## names.descript <- names.line[length(names.line)]
    ## var.names <- unlist(strsplit(names.descript, split="/&/"))
    cat("extractGatedData: FCS object=", x.name, "\n")
    cat("                  history= ", hist.str, "\n")
    cat("                  position of current gate index in x@gate=", gateNum, "\n")
    cat("                  value of index for inclusion (IndexValue.In)=", IndexValue.In, "\n\n")
  }
  
  if (is.null(x@current.data.obs)){
    stop("Cannot extractGatedData: x@current.data.obs is null.")
  }
  
  if (length(x@current.data.obs) != (dim.FCS(x)[1])){
    stop(paste("Cannot extractGatedData: Mismatch of x@current.data.obs=",
               x@current.data.obs, "with dim.FCS(x)[1]=", dim.FCS(x)[1]))
  }
  ## get the gate
  gate.idx <- x@gate[,gateNum][x@current.data.obs]
  ##  subset the matrix
  
  ## 3/14/04  Something is not right here...
  if ((dim.FCS(x)[1]) != length(gate.idx)){
    stop(paste("The gating index length",
               length(gate.idx),
               "mismatches with the data row length",
               dim.FCS(x)[1], sep=" "))
  }
  
  
  ## row reduce the data
  num.row <- sum(gate.idx==IndexValue.In)
  colnames.data <- colnames(x@data)
  if (num.row==0){
    x@data <- matrix()
    x@current.data.obs <- NULL
  } else {
    
    x@data <- matrix(as.vector(x@data[gate.idx==IndexValue.In,]),
                     byrow=FALSE, nrow=num.row)
    
    colnames(x@data) <- colnames.data
    x@current.data.obs <- x@current.data.obs[gate.idx==IndexValue.In]
  }
  
  
  
  ## DO NOT CHANGE x@gate!!
  ## row reduce the gate matrix, rows of gate match with rows of data
  ## so that subsequent extractGatedData can be implemented
  ##colnames.gate <- colnames(x@gate)
  ##if (num.row ==0 ){
  ##  x@gate <- matrix()
  ##} else {
  ##  x@gate <- matrix(as.vector(x@gate[gate.idx==IndexValue.In,]),
  ##                   byrow=FALSE, nrow=num.row)
  ##  colnames(x@gate) <- colnames.gate
  
  ##}
           
  ## make a note of the subset in the FCS metadata?  Yes,
  ## change the original to FALSE
  if (x[["original"]]==FALSE){
    x[["$TOT"]] <- dim.FCS(x)[1]
    x[["$PAR"]] <- dim.FCS(x)[2]
    ranges.data <- unlist(apply(x@data, 2, max))
    x[["$PnR"]] <- ranges.data
  } else {
    x[["RFACSadd>>$TOT"]] <- dim.FCS(x)[1]
    x[["RFACSadd>>$PAR"]] <- dim.FCS(x)[2]
    
    ## column information does not change
    ## only subset the rows
    ranges.data <- unlist(apply(x@data, 2, max))
    x[["RFACSadd>>$PnR"]] <- ranges.data
    x[["RFACSadd>>$PnS"]] <- x[["$PnS"]]
    x[["RFACSadd>>$PnN"]] <- x[["$PnN"]]
    
    x[["original"]] <- FALSE
  }
  gateName <- colnames(x@gate)[gateNum]
  x@extractGatedData.msg[gateNum] <- paste("extractGatedData:(gateNum,gateName)=(",
                                           gateNum, ",", gateName, ")",
                                           "///IndexValue.In=", IndexValue.In,
                                           sep="")
  
  return(x)
}
            
## To extract information from the history of the FCSgate object
##if (!isGeneric("extractGateHistory")) {
##  setGeneric("extractGateHistory",
##             function(x, gateNum){
##               standardGeneric("extractGateHistory")
##             })
##}

## output this: info="type",
## "data.colpos", "data.colnames", "comment",
## "prev.gateNum", "prev.gateName", "gatingrange", "IndexValue.In"
"extractGateHistory" <- function(x,gateNum){
            ## check to see if gateNum is valid
            ncol.gate <- dim(x@gate)[2]
            if ( ! (gateNum %in% 1:ncol.gate)){
              stop(paste("Input 'gateNum' is not a valid column position such as <",
                         paste(1:ncol.gate, collapse=","), "> in 'x@gate'", sep=""))
            }
            gateName <- colnames(x@gate)[gateNum]
            ## getting the strings
            hist.strings <- unlist(strsplit(x@history[gateNum], split="///"))
            type.data.str <- hist.strings[1]
            comment.str <- hist.strings[2]
            prev.gate.str <- hist.strings[3]
            gatingrange.str <- hist.strings[4]
            IndexValue.In.str <- hist.strings[5]
            ## getting the type
            type <- unlist(strsplit(unlist(strsplit(type.data.str, split="="))[2], split="\\("))[1]

            ## getting the data.colpos
            
            if (type=="uniscut"){
              data.colpos.str1 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                   split="="))[3],
                                                                   split=","))[1], split="\\("))[2]
              ## geting the data.colnames
              data.colnames.str1 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                     split="="))[3],
                                                                     split=","))[2], split="\\)"))[1]
              data.colpos <- as.numeric(data.colpos.str1)
              data.colnames <- data.colnames.str1
              if (!is.na(data.colnames)){
                if (data.colnames=="NA"){
                  data.colnames <- NA
                }
              }
            } else {
              data.colpos.str1 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                   split="="))[3],
                                                                   split=","))[1], split="\\("))[3]
              data.colpos.str2 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                   split="="))[3],
                                                                   split=","))[2], split="\\)"))[1]
              ## geting the data.colnames
              data.colnames.str1 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                     split="="))[3],
                                                                     split=","))[3], split="\\("))[2]
              data.colnames.str2 <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                                     split="="))[3],
                                                                     split=","))[4], split="\\)"))[1]
              data.colpos <- as.numeric(c(data.colpos.str1, data.colpos.str2))
              if (!is.na(data.colnames.str1)){
                if (data.colnames.str1=="NA"){
                  data.colnames.str1 <- NA
                }
              }
              if (!is.na(data.colnames.str2)){
                if (data.colnames.str2=="NA"){
                  data.colnames.str2 <- NA
                }
              }
              data.colnames <- c(data.colnames.str1, data.colnames.str2)
            }
      
            ## getting the comment
            comment <- unlist(strsplit(comment.str, split="="))
            if (length(comment)==1){
              comment <- ""
            } else {
              comment <- paste(comment[-1], collapse="=", sep="")
            }

            ## getting the prev.gateNum
            prev.gate.str <- unlist(strsplit(prev.gate.str, split="="))[2]
            prev.gateNum <- as.numeric(unlist(strsplit(unlist(strsplit(prev.gate.str, split=","))[1], split="\\{"))[2])
            
            ## getting the prev.gateName
            prev.gateName <- unlist(strsplit(unlist(strsplit(prev.gate.str, split=","))[2], split="\\}"))[1]
            if (!is.na(prev.gateName)){
              if (prev.gateName=="NA"){
                prev.gateName <- NA
              }
            }
            
            ## getting the gatingrange
          
            if (type=="uniscut"){
              gatingrange <- as.numeric(unlist(strsplit(gatingrange.str, split="="))[2])
            } else if (type=="biscut"){
              v1 <- unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[1], split="="))[2]
              v2 <- unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[2], split="="))[2]
              gatingrange <- as.numeric(c(v1,v2))
            } else if (type=="bidcut"){
             v1.1 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[1], split=","))[1], split="="))[2]
             v1.2 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[1], split=","))[2], split="="))[2]

             v2.1 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[2], split=","))[1], split="="))[2]
             v2.2 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str, split=";"))[2], split=","))[2], split="="))[2]
             gatingrange <- as.numeric(c(v1.1, v1.2, v2.1, v2.2))

            } else if (type=="bipcut"){
              v1 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str,
                                                                    split=";"))[1],
                                                    split="="))[2], split=","))
              v2 <- unlist(strsplit(unlist(strsplit(unlist(strsplit(gatingrange.str,
                                                                    split=";"))[2],
                                                    split="="))[2], split=","))
              gatingrange <- as.numeric(c(v1, v2))

            }
             
              
             ## getting biscut.quadrant
            if (type!="biscut"){
              biscut.quadrant <- NULL
            } else {
              biscut.quadrant <-  unlist(strsplit(unlist(strsplit(unlist(strsplit(type.data.str,
                                                                            split="="))[3],
                                                            split=","))[5], split="\\)"))[1]
            }
              
            ## getting the current IndexValue.In
            IndexValue.In <- as.numeric(unlist(strsplit(IndexValue.In.str, split="="))[2])

            ####### NOTE: ####################################
            ## extracted before?
            ## if there has been an extraction, then we assign IndexValue.In to that what was extracted

            if (x@extractGatedData.msg[gateNum] != "NONE"){
              IndexValue.In <- as.numeric(unlist(strsplit(unlist(strsplit(x@extractGatedData.msg[gateNum],
                                                                          split="///"))[2], split="="))[2])
            }
            
            return(list("gateNum"=gateNum, "gateName"=gateName,
                        "type"=type,
                        "biscut.quadrant"=biscut.quadrant, 
                        "data.colpos"=data.colpos,
                        "data.colnames"=data.colnames,
                        "IndexValue.In"=IndexValue.In,
                        "gatingrange"=gatingrange,
                        "prev.gateNum"=prev.gateNum,
                        "prev.gateName"=prev.gateName,
                        "comment"=comment))
                        
          }

###################################################################
##  PLOTTING FUNCTIONS
###################################################################

###  STUFF TO DO: 3/11/2004-3/12/2004
###  Interactive Gating: iGate is S4??? YES, FHCRC.HVTNFCS, VRC.HVTNFCS are S3
###  that use createGate, extractGatedData, iGate

##if (!isGeneric("icreateGate")) {
##  setGeneric("icreateGate",
##             function(x,...) {
##               standardGeneric("icreateGate")
##             })
##}


               ##  

"icreateGate" <- function(x,
                          varpos=NULL,
                          gatingrange=NULL,
                          type=NULL,
                          biscut.quadrant=NULL,
                          prev.gateNum=NULL,
                          prev.IndexValue.In=NULL,
                          comment=NULL,
                          pchtype=".",
                          ## STOPPED Need to revise
                          ## the signature same as 'createGate' 
                          MY.DEBUG=TRUE,
                          prompt.all.options=TRUE) {

    ##  This is an interactive gating and plotting function
  
  ##  For a sequential gating scheme please use
  ##  FHCRC.HVTNFCS or VRC.HVTNFCS, which are
  ##  the gating schemes that are used, respectively,
  ##  in Julie McElrath's lab and Mario Roederer's Lab
  
  ## IF S3 object (will phase out)
  ##  For basic gating use: gate.FCS
  ##  For basic subsetting use: subset.FCS
  ##  For basic plotting use: plotvar.FCS.FCS & showgate.FCS
  ## subset.colNum is the position of the subsetting index variable

            
  ## IF S4 object (working on right now)
  ## basic creating gating index: createGate
  ## basic subseting of rows: extractGatedData
  ## basic plotting: (still in progress)
  
  ## IF ANY NULL VARIABLE IN THE SIGNATURE IS NOT NULL,
  ## THE VALIDITY OF THE INPUT VARIABLE WILL BE CHECKED
  ## IF VALID, THEN THE VALUE WILL BE USED, OTHERWISE
  ## THE USER WILL BE PROMPTED FOR CORRECT INPUT
  
  
  
  ##------------------------
  ## New S4 stuff begins
  ##------------------------
  if (MY.DEBUG){
    print(paste("icreateGate: FCS object =", x[["objectname"]], sep=" "))
  }
  names.orig <- colnames(x@data)
  n.row <- dim.FCS(x)[1]
  n.col <- dim.FCS(x)[2]
  
  ##---------------------
  ## prompt for varpos
  ##---------------------
  num.gtparms <- NULL
  if (!is.null(type)){
    if (type=="uniscut"){
      num.gtparms <- 1
    }
    if (type %in% c("biscut", "bidcut", "bipcut")){
      num.gtparms <- 2
    }
  } 
  while(is.null(num.gtparms)){
    print("How many parameters are to be used in the gate?")
    print("  Input 1 or 2.")
    num.gtparms <- scan(n=1)
    if (!(num.gtparms %in% c(1,2))){
      num.gtparms <- NULL
    }
  }
  
            
  if (is.null(varpos)){
    varpos <- ""
    while((!is.numeric(varpos)) || (!(varpos %in% 1:n.col))){
      print("Prompt 'varpos':")
      
      if ((num.gtparms %in% c(1,2))){
        verb.1 <- ifelse(num.gtparms==1, "is", "are")
        msg.1 <- ifelse(num.gtparms==1, "Input a single position value.",
                        "Input two position values; each value followed by the 'enter' key.")
        print(paste("What", verb.1," is the column parameter position(s) of the data to be gated?", sep=""))
        print(paste("  ", msg.1, sep=""))
        print(paste("  names of the columns=", paste(names.orig, collapse=",")))
        
        varpos <- scan(n=num.gtparms)
        varpos <- unlist(varpos)
        if (!(length(varpos) %in% c(1,2))){
          varpos <- NULL
        }
      } else {
        varpos <- ""
      }
      
    }
  }
  ##--------------------------------
  ## Prompt for prev.gateNum
  ##--------------------------------
  if (is(x, "FCSgate")){
    if (!is.null(x@gate)){
      no.prevgate <- FALSE
      prev.colpos <- 1:(dim(x@gate)[2])
      if (is.null(prev.gateNum)){
        prev.gateNum <- ""
        while (!no.prevgate &
               !(prev.gateNum %in% c(0, prev.colpos))) {
          print("Prompt 'prev.gateNum':")
          print("What is the previous gate index (ie, the column parameter position of x@gate)?")
          print(paste("  Input 0 (no previous gate) or a value: ", paste(prev.colpos, collapse=","), sep=""))
          
          prev.gateNum <- scan(n=1)
          
          if (prev.gateNum==0){
            no.prevgate <- TRUE
          }
        }
      }
      
    } else {
      if (!is.null(prev.gateNum)){
        print("Prompt 'prev.gateNum':")
        print("       'x@gate' is NULL.")
        print("       The 'prev.gateNum' is set to NULL.")
        prev.gateNum <- NULL
      }
      
    }
    
  } else {
    
    if (!is.null(prev.gateNum)){
      print("Prompt 'prev.gateNum':")
      print("       Input 'x' is not of class 'FCSgate'.")
      print("       The 'prev.gateNum' is set to NULL.")
      prev.gateNum <- NULL
    }
    
  }
  ##----------------------------------------------
  ## Prompt for the IndexValue.In for prev.gateNum (if applicable)
  ##----------------------------------------------
  
  ## Account for previous gate so subset the data
  
  if (!is.null(prev.gateNum)){
    ## find the IndexValue.In for previous gate
    if (is.null(prev.IndexValue.In)){
      prev.IndexValue.In <- ""
      prev.idx <- x@gate[,prev.gateNum]
      prev.rng <- range(prev.idx, na.rm=TRUE)
      while (!(prev.IndexValue.In %in% prev.rng[1]:prev.rng[2]) || !is.numeric(prev.IndexValue.In)){
        print("Prompt 'prev.IndexValue.In':")
        print("       What is the inclusion value for the previous index?")
        print(paste("       Input a single value:", paste(prev.rng[1]:prev.rng[2], collapse=","), sep=""))
        
        prev.IndexValue.In <- scan(n=1)
      }
    }
    ## obtain the extracted
    x.ext <- extractGatedData(x,gateNum=prev.gateNum, IndexValue.In=prev.IndexValue.In)
    
  } else {
    x.ext <- x
  }


            
  ##---------------------
  ## prompt for type
  ##---------------------
  type.gates <- c("uniscut", "biscut", "bipcut", "bidcut")
  if (is.null(type)){
    type <- ""
  }
  while((!is.character(type)) || (!(type %in% type.gates))){
    print("Prompt 'type':")
    print("What is the type of gating to be desired?")
    print("  Input one of the following:")
    if (length(varpos)==1){
      print("     uniscut = univariate single cut")
    } else if (length(varpos)==2){
      print("     biscut = bivariate single cut")
      print("     bidcut = bivariate double cut")
      print("     bipcut = bivariate polygonal cut")
    }
    
    type <- scan(n=1, what="")
    if (length(varpos)==1 & type!="uniscut"){
      type <- NULL
    }
    if (length(varpos)==2 & (!(type %in% c("biscut", "bipcut", "bidcut")))){
      type <- NULL
    }
    
  }

  if (prompt.all.options){
    ## Getting the title of the plot
    print("What title would you like on your plot? Please enclose in quotes.")
    input.title <- scan(n=1, what="")
  } else {
    input.title <- ""
  }
  ## Getting any additional comments associated with the Index column
  ## gating parameter
  if (is.null(comment)){
    print(paste("Any additional comments to attach to this gating?",
                "If so, please enclose in quotes."))
    comment <- scan(n=1, what="")
  }
  
  ## Getting the color of the gate #########
  if (prompt.all.options){
    print(paste("What would you like the color of the gate to be?",
                "Default is 12 (purple)"))
    input.color <- scan(n=1)
  } else {
    input.color <- 12
  }
  
  ## Getting the gating range for a univariate single cut
  
  ## histogram Plotting of the univariate variable and the gate is
  ## shown
  ##-----------------------------
  ## Display ContourScatterPlot
  ##-----------------------------
  
  getCSPlot.info <- function() {
    ## AJR: what does this do?
    ## prompts for information about the plot options
    ## we can also use the default options
    
    get.grid <- function(word1, word2) {
      grid <- 3
      if(!is.element(grid, c(0,1))) {
        print(paste("plotvar.FCS:", word1,
                    "is seq(0, 1025, by=25). Change",  word1,
                    "? 1=yes OR 0=no, then press ENTER.", sep=" "))
        grid <- scan(n=1)
      }
      if (grid==1) {
        min.val <- 0
        max.val <- 0
        
        while(min.val>=max.val) {
          print("What is the minimum value of x?")
          min.val <- scan(n=1)
          print(paste("What is the maximum value of", word2, "?", sep=""))
          max.val <- scan(n=1)
        }
        print("What is the value of the length between grid marks?")
        between.val <- scan(n=1)
      } else {
        if (grid==0) {
          min.val <- 0
          max.val <- 1025
          between.val <- 25
        }
      }
      return(seq(min.val, max.val, by=between.val))
    }
    
    x.grid <- get.grid("x.grid", "x")
    y.grid <- get.grid("y.grid", "y")
    
    colorType <- ""
    while(!is.element(colorType,
                      c("heat", "terrain", "rainbow", "cm", "gray"))) {
      print(paste("What image color would you like",
                  "(heat, terrain, rainbow, cm, gray)?",
                  "Please make sure the input is in quotes and press ENTER."))
      print("(NOTE: Some systems will only allow gray)")
      colorType <- scan(n=1, what="")
      print(paste("How many colors would you like?",
                  "Try 2 if not sure;",
                  "there will be an error if using more colors than displayable."))
      numCol <- scan(n=1)
    }
    
    ##    if (colorType=="heat") {
    ##      image.col <- heat.colors(numCol)
    ##    } else if (colorType=="terrain"){
    ##      image.col <- terrain.colors(numCol)
    ##    } else if (colorType=="rainbow"){
    ##      image.col <- rainbow(numCol)
    ##    } else if (colorType=="cm") {
    ##      image.col <- cm.colors(numCol)
    ##    } else if (colorType=="gray"){
    ##      image.col <- gray(0:numCol/numCol)
    ##    }
    
    image.col <- switch(colorType,
                        heat=heat.colors(numCol),
                        terrain=terrain.colors(numCol),
                        rainbow=rainbow(numCol),
                        cm=cm.colors(numCol),
                        gray= gray(0:numCol/numCol))
    
    print(paste("What number of levels for the contour plot?",
                "Default is 50, but try 25 if not sure."))
    numlev <- scan(n=1)
    
    return(list(x.grid=x.grid, y.grid=y.grid,
                image.col=image.col, numlev=numlev))
  }
  
  ## end of getCSPlot.info
  
  if (length(varpos)==1){
    type.plotvar <- "uni"
    plotType.plotvar <- "hist"
  }
  if (length(varpos)==2){
    type.plotvar <- "bi"
    plotType.plotvar <- "ContourScatterPlot"
    
  }
  
  ##--------------------------------
  ## Prompt for gatingrange
  ##--------------------------------
  
  if (type=="uniscut") {
    plot.again <- TRUE
    while(plot.again==TRUE) {
      print("   plotvar.FCS: Making univariate histogram; Please Wait...")
      
      plotvar.FCS(x.ext,varpos, type=type.plotvar, plotType=plotType.plotvar,
                  names.var=colnames(x.ext@data)[varpos],title.pl=input.title)
      
      if (is.null(gatingrange)){
        lb <- ""
        while ( !(is.numeric(lb)) || !(lb %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]) )){
          print("What is the lower bound cut-off?")
          
          lb <- scan(n=1)
        }
        gatingrange <- lb
      } 

      if (prompt.all.options){
        IndexValue.In <- NULL
        if (is.null(IndexValue.In)){
          IndexValue.In <- ""
          while( !(is.numeric(IndexValue.In)) || !(IndexValue.In %in% c(0,1) )){
            print(paste("Selection: 0=We will select observations < the lower bound cut-off,", gatingrange))
            print(paste("           1=We will select observations >= the lower bound cut-off,", gatingrange))
            IndexValue.In <- scan(n=1)
            
          }
        }
      } else {
        IndexValue.In <- 1
      }
      if (MY.DEBUG){
        print("   createGate: Making selection/gate index; Please Wait...")
      }
      x.ext <- createGate(x.ext, varpos=varpos, gatingrange=gatingrange, type=type)
      curr.gt <- dim(x.ext@gate)[2]
      if (MY.DEBUG){
        print("   showgate.FCS: Showing gate on plot; Please Wait...")
      }
      showgate.FCS(as(x.ext@data[,varpos], "matrix"), gatingrange=gatingrange,
                   Index=x.ext@gate[x.ext@current.data.obs,curr.gt],
                   type="uniscut", IndexValue.In=IndexValue.In,
                   coltype=input.color, pchtype=pchtype)
      
      if (prompt.all.options){
        print("Redo gate and plot again?  1=TRUE, 0=FALSE.")
        plot.again <- scan(n=1)
        if (plot.again) {
          ## set values to null
          gatingrange <- NULL
        }
      } else {
        plot.again <- FALSE
      }
      
    }
    
    ## Getting the gating range of a bivariate double cut
    
    ## Bivariate ContourScatterPlot
    
    ## Plot and show the Gate results
    
  }

  if (type=="biscut") {
   
    plot.again <- TRUE
    while(plot.again==TRUE) {
##########  Initializing options for the CSP plot #################
      
      if (prompt.all.options){
        CSP.info <- getCSPlot.info()
      } else {
        CSP.info=list(x.grid=seq(0, 1025, by=25),
          y.grid=seq(0, 1025, by=25),
          image.col=heat.colors(2), numlev=25)
      }
      if (MY.DEBUG){
        print("   plotvar.FCS: Making bivariate contour plot; Please Wait...")
      }
      plotvar.FCS(x.ext,varpos, type=type.plotvar, plotType=plotType.plotvar,
                  names.var=colnames(x.ext@data)[varpos],title.pl=input.title,
                  hexbin.CSPlot=FALSE, x.grid.CSPlot=CSP.info$x.grid,
                  y.grid.CSPlot=CSP.info$y.grid,
                  image.col.CSPlot=CSP.info$image.col,numlev.CSPlot=CSP.info$numlev)
      
      names.datavar <- colnames(x.ext@data)[varpos]
      if (is.null(gatingrange)){
        xbound <- ""
        while ( !is.numeric(xbound) || !(xbound %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the cut-off for the x-variable,",
                      names.datavar[1], "?", sep=" "))
          xbound<- scan(n=1)
        }
        
        ybound <- ""
        while ( !is.numeric(ybound) || !(ybound %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the cut-off for the y-variable,",
                      names.datavar[2], "?", sep=" "))
          ybound <- scan(n=1)
          
        }
        
        gatingrange <- c(xbound, ybound)
      }
      
       ##abline(h=ybound, col=input.color)
       ##abline(v=xbound, col=input.color)
      if (is.null(biscut.quadrant)==TRUE) {
        
        biscut.quadrant <- ""
        while (charmatch(biscut.quadrant,
                         c("+/+", "-/-", "-/+", "+/-"))==0){
          print(paste("Input the quadrant to be selected",
                      "(enclosed in quotes)::",
                      "+/+, +/-, -/-, -/+"))
          biscut.quadrant <- scan(n=1, what=character(0))
        }
      }
      if (MY.DEBUG){
        print("   createGate: Making selection/gate index; Please Wait...")
      }
      x.ext <- createGate(x.ext, varpos=varpos, gatingrange=gatingrange, type=type,
                          biscut.quadrant=biscut.quadrant)
      
      curr.gt <- dim(x.ext@gate)[2]
      if (MY.DEBUG){
        print("   showgate.FCS: Showing gate on plot; Please Wait...")
      }
      showgate.FCS(x.ext@data[,varpos], gatingrange=gatingrange,
                   Index=x.ext@gate[x.ext@current.data.obs,curr.gt],
                   type=type, biscut.quadrant=biscut.quadrant,IndexValue.In=1,
                   coltype=input.color,pchtype=pchtype)
      if (prompt.all.options){
        print("Redo gate and plot again?  1=TRUE, 0=FALSE.")
        plot.again <- scan(n=1)
        if (plot.again) {
          ## set values to null
          gatingrange <- biscut.quadrant<- NULL
        }
      } else {
        plot.again <- FALSE
      }
      
    }
  }
  if (type=="bidcut") {
    plot.again <- TRUE
    while(plot.again==TRUE) {
      
##########  Initializing options for the CSP plot #################
      
      if (prompt.all.options){
        CSP.info <- getCSPlot.info()
      } else {
        CSP.info=list(x.grid=seq(0, 1025, by=25),
          y.grid=seq(0, 1025, by=25),
          image.col=heat.colors(2), numlev=25)
      }
      if (MY.DEBUG){
        print("   plotvar.FCS: Making bivariate contour plot; Please Wait...")
      }
       
     
      plotvar.FCS(x.ext,varpos, type=type.plotvar, plotType=plotType.plotvar,
                  names.var=colnames(x.ext@data)[varpos],title.pl=input.title,
                  hexbin.CSPlot=FALSE, x.grid.CSPlot=CSP.info$x.grid,
                  y.grid.CSPlot=CSP.info$y.grid,
                  image.col.CSPlot=CSP.info$image.col,numlev.CSPlot=CSP.info$numlev)
  
###########  stopped here #####################
      if (is.null(gatingrange)){
        xlow <- ""
        names.datavar <- colnames(x.ext@data)[varpos]
        while (!is.numeric(xlow) || !(xlow %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the lower bound cut-off for the x-variable,",
                      names.datavar[1], "?", sep=" "))
          xlow <- scan(n=1)
        }
        
        xhigh <- ""
        while (!is.numeric(xhigh) || !(xhigh %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the upper bound cut-off for the x-variable,",
                      names.datavar[1], "?", sep=" "))
          xhigh <- scan(n=1)
        }
        ylow <- ""
        while (!is.numeric(ylow) || !(ylow %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the lower bound cut-off for the y-variable,",
                      names.datavar[2], "?", sep=" "))
          ylow <- scan(n=1)
        }
        
        yhigh <- ""
        while (!is.numeric(yhigh) || !(yhigh %in% (range(x.ext@data)[1]):(range(x.ext@data)[2]))){
          print(paste("What is the upper bound cut-off for the y-variable,",
                      names.datavar[2], "?", sep=" "))
          yhigh <- scan(n=1)
        }
        
        
        gatingrange <- c(xlow, xhigh, ylow, yhigh)
      }
      if (MY.DEBUG){
        print("   createGate: Making selection/gate index; Please Wait...")
      }
      x.ext <- createGate(x.ext, varpos=varpos, gatingrange=gatingrange, type=type)
      
      curr.gt <- dim(x.ext@gate)[2]
      if (MY.DEBUG){
          print("   showgate.FCS: Showing gate on plot; Please Wait...")
      }
      
      
      showgate.FCS(x.ext@data[,varpos], gatingrange=gatingrange,
                   Index=x.ext@gate[x.ext@current.data.obs,curr.gt],
                   type=type, IndexValue.In=1,
                   coltype=input.color, pchtype=pchtype)
      
      if (prompt.all.options){
        print("Redo gate & plot again?  1=TRUE, 0=FALSE.")
        plot.again <- scan(n=1)
        if (plot.again) {
          ## set values to null
          gatingrange <- NULL
        }
      } else {
        plot.again <- FALSE
      }
    }
    
    ## Getting the gating range of a bivariate polygonal cut
    
    ## Bivariate ContourScatterPlot Plotting and showing of gate
    ## results
    
  }
  if (type=="bipcut") {
    
    lib.there <- require(splancs)
    if (lib.there==FALSE) {
      stop("Library splancs can not be found but is needed")
    }
    plot.again <- TRUE
    while(plot.again==TRUE) {
      if (prompt.all.options){
        CSP.info <- getCSPlot.info()
         
      } else {
        CSP.info=list(x.grid=seq(0, 1025, by=25),
          y.grid=seq(0, 1025, by=25),
          image.col=heat.colors(2), numlev=25)
      }
     
      if (MY.DEBUG){
        print("   plotvar.FCS: Making bivariate contour plot; Please Wait...")
      }
      plotvar.FCS(x.ext,varpos, type=type.plotvar, plotType=plotType.plotvar,
                  names.var=colnames(x.ext@data)[varpos],title.pl=input.title,
                  hexbin.CSPlot=FALSE, x.grid.CSPlot=CSP.info$x.grid,
                  y.grid.CSPlot=CSP.info$y.grid,
                  image.col.CSPlot=CSP.info$image.col,numlev.CSPlot=CSP.info$numlev)

      if (is.null(gatingrange)){
        print(paste("On the graph,",
                    "left click the vertices of your polygon boundary.",
                    "right click when done.", sep=" "))
        linesToDraw <- locator()
        numberOfPoints <- length(linesToDraw$x)
        gatingrange <- c(linesToDraw$x[c(1:numberOfPoints,1)],
                         linesToDraw$y[c(1:numberOfPoints,1)])
      }
      if (MY.DEBUG){
        print("   createGate: Making selection/gate index; Please Wait...")
      }
      x.ext <- createGate(x.ext, varpos=varpos, gatingrange=gatingrange, type=type)
      
      curr.gt <- dim(x.ext@gate)[2]
      if (MY.DEBUG){
        print("   showgate.FCS: Showing gate on plot; Please Wait...")
      }
      showgate.FCS(x.ext@data[,varpos], gatingrange=gatingrange,
                   Index=x.ext@gate[x.ext@current.data.obs,curr.gt],
                   type=type, IndexValue.In=1,
                   coltype=input.color, pchtype=pchtype)
      if (prompt.all.options){
        print("Redo gate & plot again? 1=TRUE, 0=FALSE")
        plot.again <- scan(n=1)
        if (plot.again) {
          ## set values to null
          gatingrange <- NULL
        }
      } else {
        plot.again <- FALSE
      }
      print(CSP.info)
    }
  }
  if (MY.DEBUG){
    print("   createGate: Finalizing output selection/gate index; Please Wait...")
    print("")
  }
  x <- createGate(x,varpos=varpos,
                  gatingrange=gatingrange,
                  type=type, biscut.quadrant=biscut.quadrant,
                  prev.gateNum=prev.gateNum,
                  prev.IndexValue.In=prev.IndexValue.In,
                  comment=comment, MY.DEBUG=MY.DEBUG)
  x
  
}


"showgate.FCS" <-  function(data.mat,
                            gatingrange,
                            Index, 
                            type=c("uniscut",
                              "biscut",
                              "bidcut",
                              "bipcut"),
                            IndexValue.In=1,
                            coltype=12,
                            pchtype=8,
                            biscut.quadrant=c("+/-",
                              "-/-", "+/+", "-/+")){
######################
  # input
  #####################

  
  # type        type of cut/gating
  #              uniscut  univariate single cut
  #              bidcut bivariate double cut
  #              bipcut bivariate polygonal cut
  
  # gatingrange gating threshold range (must contain all points in the range) ie, univariate c(x1,x2),
  #             bivariate c(x1,x2, y1,y2), or polygonal thresholds (n-sided polygon)
  #             (a vector of vectors ie,cbind(c(x1, x2, ...,xn, x1), c(y1, y2, ...,yn, y1))
  #             which denotes the outer points of the polygon vertices)

  # coltype          the color of the gate
  
  # data.bipcut         ONLY FOR POLYGONAL GATE (type="bipcut"): the two column variables:
  #                            a data matrix (m X 2) where m is the number of cells/rows
  # pchtype.bipcut      ONLY FOR POLYGONAL GATE (type="bipcut"): an option for the
  #                            point size of the data point outside the polygon gate
  # coltype.bipcut      ONLY FOR POLYGONAL GATE (type="bipcut"): an option for the
  #                            color of the data point outside the polygon gate
  #             
  #######################
  # output
  #####################
  # plots the gates

  if (!is(data.mat, "matrix")){
    data.mat <- as(data.mat, "matrix")
  }
  
  type <- match.arg(type)
  lib.there <- require(splancs)
  
  if(lib.there==FALSE) {
    stop("The library splancs is needed and could not be found")
  }

  if (!((dim(data.mat)[2]) %in% c(1,2))) {
    stop("Input 'data.mat' should have 1 or 2 columns.")
  }
  
   
  if (type=="uniscut"){
    
    abline(v=gatingrange, col=coltype)
   
    
  }
  
  if (type=="bidcut"){
    xlow <- min(gatingrange[1:2])
    xhigh <- max(gatingrange[1:2])
    ylow <- min(gatingrange[3:4])
    yhigh <- max(gatingrange[3:4])
   
    abline(h=ylow, col=coltype)
    abline(h=yhigh, col=coltype)
    abline(v=xlow, col=coltype)
    abline(v=xhigh, col=coltype)

   
  }

  if (type=="biscut" ){
    xbound <- gatingrange[1]
    ybound <- gatingrange[2]
    abline(h=ybound, col=coltype)
    abline(v=xbound, col=coltype)
    biscut.quadrant <- match.arg(biscut.quadrant)


  }
  



  if (type!="uniscut"){
    
    data.selected <- data.mat[Index==IndexValue.In, 1:(dim(data.mat)[2])]
    points(data.selected, col=coltype, pch=pchtype)
  }
}


  
 
#######################################################################
##  gate.IPC may not work...
#######################################################################


"gate.IPC" <- function(myFCSobj,
                       var.gate,
                       var.pos=1:(dim(myFCSobj@data)[2]),
                       num.bins=10,
                       joint=FALSE,
                       range.var=range(myFCSobj@data[,var.pos]),
                       break10 =seq(range.var[1]-1, range.var[2],
                         by=range.var[2]/num.bins),
                       title="",use.shortnames=FALSE,
                       color.image=gray((25:5/25)[-c(1,2,3, 4, 5, 6)]), #rainbow(16, start=.4, end=.1),#
                       xwidth.scale=5, ntrans=1,
                       hist.plotted=FALSE,
                       image.plotted=TRUE,
                       para.plotted=FALSE,
                       lines.plotted=TRUE,
                       legend.plotted=TRUE,
                       lwd.vec=1:7,
                       lty.vec=rep(1,7),
                       col.vec=7:1,
                       range.image=c(0, dim(myFCSobj@data)[1]),
                       shrink.legend=TRUE,
                       horizontal.legend = TRUE,
                       offset.legend=0.03,
                       nlevel.legend=length(color.image),
                       xlab.image="",
                       ylab.image="Bins",
                       MY.DEBUG=FALSE,...) {

  ## also check to see if it is in range

  if (!checkvars(myFCSobj, MY.DEBUG=MY.DEBUG)){
    myFCSobj <- fixvars(myFCSobj, MY.DEBUG=MY.DEBUG)
  }
  
  if(joint==TRUE) {
    plottype <- JointImageParCoord
  } else {
    plottype <- ImageParCoord
  }
  ## preliminary plots
   info <- plottype(myFCSobj@data[,var.pos],
                    num.bins=num.bins,
                    range.var=range.var,
                    break10 =break10,
                    title=title,
                    use.shortnames=use.shortnames,
                    color.image=color.image,
                    xwidth.scale=xwidth.scale,
                    ntrans=ntrans,
                    hist.plotted=hist.plotted,
                    image.plotted=image.plotted,
                    para.plotted=para.plotted,
                    lines.plotted=lines.plotted,
                    legend.plotted=legend.plotted,
                    lwd.vec=lwd.vec,
                    lty.vec=lty.vec,
                    col.vec=col.vec,
                    range.image=range.image,
                    shrink.legend=shrink.legend,
                    horizontal.legend = horizontal.legend,
                    offset.legend=offset.legend,
                    nlevel.legend=nlevel.legend,
                    xlab.image=xlab.image,
                    ylab.image=ylab.image,... )
  ##plot on which to gate
  info <- plottype(myFCSobj@data[,var.pos],
                   num.bins=num.bins,
                   range.var=range.var,
                   break10 =break10,
                   title=title,
                   use.shortnames=use.shortnames,
                   color.image=color.image,
                   xwidth.scale=xwidth.scale,
                   ntrans=ntrans,
                   hist.plotted=FALSE,
                   image.plotted=TRUE,
                   para.plotted=FALSE,
                   lines.plotted=FALSE,
                   legend.plotted = FALSE,
                   lwd.vec=lwd.vec,
                   lty.vec=lty.vec,
                   col.vec=col.vec,
                   range.image=range.image,
                   shrink.legend=shrink.legend,
                   horizontal.legend = horizontal.legend,
                   offset.legend=offset.legend,
                   nlevel.legend=nlevel.legend,
                   xlab.image=xlab.image,
                   ylab.image=ylab.image,... )

 
  prompt.it <- function(){
    print(paste("On the graph, left click the y-axis",
                "boundary at which the observations",
                "above and including the boundary",
                "are selected with the inclusion index value=1", sep=" "))
    print(paste("for variable", var.gate, ":", colnames(myFCSobj@data)[var.gate],sep=" "))
    point.bd<- locator(n=1)
    return(point.bd)
  }
  
  bin.bd <- -99999
  while (is.element(bin.bd, 1:length(info$breaks))==FALSE) {
    point.bd <- prompt.it()
    bin.bd <-trunc(point.bd$y)
  }
  
  ##gating only one variable, var.gate
  #sort by the first variable


  ######################  stopped here: do we gate here?  how do it in S4?   ############################
 ## var.sorted<- sort(myFCSobj@data[,var.gate],
  ##                  decreasing=FALSE, index.return=TRUE)
  ##  var.sorted.pos <- var.sorted$ix
  
  ## myFCSobj$data <- myFCSobj$data[var.sorted.pos,] 
  
  ## breaks <- info$breaks
  ## convert the bin.bd to a variable value
  ## how to do this?
  
  ## we are going to include everything in the bin that is clicked
  ##  cutoff <- breaks[bin.bd+1]
  ## is the cutoff right? yes
  
  ## up to an including the cutoff
  ## pos.st1 <- which(var.sorted$x>=cutoff)
  ##  plot1.pos <-var.sorted.pos[pos.st1] 
  ##  pos.st2 <- which(var.sorted$x<cutoff)
  ##  plot2.pos <- var.sorted.pos[pos.st2]
  
  ## univariate gating
  myFCSobj <- createGate(myFCSobj, varpos=var.gate, gatingrange=break10[bin.bd],
                        type="uniscut")

  ## observations >= cuttoff/gatingrange
  x.gated1 <- extractGatedData(myFCSobj,
                               gateNum=dim(myFCSobj@gate)[2],
                               IndexValue.In=1)

  ## observations < cutoff/gatingrange
   x.gated0 <- extractGatedData(myFCSobj,
                                gateNum=dim(myFCSobj@gate)[2],
                                IndexValue.In=0)

  tot.var <- dim.FCS(myFCSobj)[2]
  info.sub1 <- plottype(x.gated1@data[,var.pos],
                        num.bins=num.bins,range.var=range.var, 
                        break10 =break10,
                        title=paste("Subplot 1:", title, sep=" "),
                        use.shortnames=use.shortnames,
                        color.image=color.image,
                        xwidth.scale=xwidth.scale,
                        ntrans=ntrans,
                        hist.plotted=hist.plotted,
                        image.plotted=image.plotted,
                        para.plotted=para.plotted,
                        lines.plotted=lines.plotted,
                        legend.plotted = legend.plotted,
                        lwd.vec=lwd.vec,
                        lty.vec=lty.vec,
                        col.vec=col.vec,
                        range.image=range.image,
                        shrink.legend=shrink.legend,
                        horizontal.legend = horizontal.legend,
                        offset.legend=offset.legend,
                        nlevel.legend=nlevel.legend,
                        xlab.image=xlab.image,
                        ylab.image=ylab.image,... )

  info.sub2 <- plottype(x.gated0@data[,var.pos],
                        num.bins=num.bins,
                        range.var=range.var,
                        break10 =break10,
                        title=paste("Subplot 2:", title, sep=" "),
                        use.shortnames=use.shortnames,
                        color.image=color.image,
                        xwidth.scale=xwidth.scale,
                        ntrans=ntrans,
                        hist.plotted=hist.plotted,
                        image.plotted=image.plotted,
                        para.plotted=para.plotted,
                        lines.plotted=lines.plotted,
                        legend.plotted = legend.plotted,
                        lwd.vec=lwd.vec,
                        lty.vec=lty.vec,
                        col.vec=col.vec,
                        range.image=range.image,
                        shrink.legend=shrink.legend,
                        horizontal.legend = horizontal.legend,
                        offset.legend=offset.legend,
                        nlevel.legend=nlevel.legend,
                        xlab.image=xlab.image,
                        ylab.image=ylab.image,...)

  return(list(info.total=info,
              info.sub1=info.sub1,
              info.sub2=info.sub2, 
              FCSgateobj=myFCSobj ))
}


