pkci2.flowcytest <- function(controldata, stimuldata,crit=.999,alpha=.05){

  ##
  ## Function to calculate gate (k.hat), percent-stimulated cells, and a 
  ## 95% CI to test for a sig. diff. in proportion of stimulated cells and
  ## control cells above the gate
  ## 
  ##

  ## data: vector of control cells (first n1 elements) and stim cells
  ##       (next n2 elements)
  ## n1: no. of cells in control sample
  ## n2: no of cells in stim sample
  ## crit: percentile of control sample to use as gate
  ## alpha: for construction of (1-alpha)% CI
  
 
  controldata <- unlist(controldata)
  stimuldata <- unlist(stimuldata)
  cat("FLOWCYTEST: PKCI2", "\n")
  n1 <- length(controldata)
  n2 <- length(stimuldata)
  ## Note: provided length(unstimuldata)*crit is integer, don't need floor
  k.hat <- sort(controldata)[floor(crit*length(controldata))]
  ps.hat <- length(stimuldata[stimuldata > k.hat])/length(stimuldata)
  pc.hat <- length(controldata[controldata > k.hat])/length(controldata)

  #  H0: ps.hat - pc.hat = 0
  #  H1.1: ps.hat - pc.hat > 0 OR ps.hat > pc.hat
  # H1.2: ps.hat - pc.hat != 0

  # alpha=Pr(reject H0 | H0 is true)
  # p.val.2sided = Pr(|Z| > z.stat )= Pr(Z > z.stat) + Pr(Z <-z.stat)
  # p.val.1sided = Pr(Z > z.stat)
  
  lb <- round(ps.hat - pc.hat - qnorm(1-alpha/2)*
              sqrt( ps.hat*(1-ps.hat)/n2 + pc.hat*(1-pc.hat)/n1), 6)

  # what about the upper bound of the CI
  ub <-  round(ps.hat - pc.hat + qnorm(1-alpha/2)
               *sqrt( ps.hat*(1-ps.hat)/n2 + pc.hat*(1-pc.hat)/n1), 6)
  
  z.stat <- ((ps.hat - pc.hat)-0)/sqrt(ps.hat*(1-ps.hat)/n2
                                       + pc.hat*(1-pc.hat)/n1)

  p.val.1sided.pkci2 <- pnorm(z.stat, lower.tail=FALSE)

  # Normal curve is symmetric
  p.val.2sided.pkci2 <- 2*(pnorm(abs(z.stat), lower.tail=FALSE)) 


 
  # One sided Test for H1.1 (1=reject H0, 0=cannot reject H0)
  test1 <- ifelse(lb > 0, 1 ,0)
  #Two sided test for H1.2(1=reject H0, 0=cannot reject H0)
  test2 <- ifelse(lb <= 0 & 0 <= ub, 0 ,1)
  
  output <- list(k.hat.pkci2=round(k.hat,6),pc.hat.pkci2=round(pc.hat,6),
                 ps.hat.pkci2=round(ps.hat,6),
                 lb.pkci2=lb, ub.pkci2=ub,
                 test.1pkci2=test1, pval1.pkci2=p.val.1sided.pkci2, test.2pkci2=test2,
                 pval2.pkci2=p.val.2sided.pkci2)

  cat("   ", "Test pkci2: Standard Normal approximation of two-sample binomial statistics", "\n")
  cat("\n")
  print(paste("k.hat,", k.hat, ",is the gate/percentile based on the control data", sep=" "))
  print(paste("   and the user specified critical proportion of, crit", sep=" "))
  print( paste(round(ps.hat, 5), ",ps.hat is the proportion of stimulated data above k.hat", sep=" "))
  print(paste(round(pc.hat,5), ", pc.hat is the proportion of the control data above k.hat,", sep=" " ))
  cat("\n")
  cat(" Null: H0: ps.hat = pc.hat OR ps.hat-pc.hat = 0", "\n")
  cat(" One-sided Alternative: H1.1: ps.hat - pc.hat > 0 OR ps.hat > pc.hat", "\n")
  cat(" Two-Sided Alternative:  H1.2: ps.hat - pc.hat != 0", "\n")

  cat("\n")
  cat(paste("Standard Normal Z Statistic:", z.stat, sep=" "), "\n")
  cat(paste("One sided p-value:", p.val.1sided.pkci2, sep=" "), "\n")
  cat(paste("Two sided p-value:", p.val.2sided.pkci2, sep=" "), "\n")
   cat("\n")
  cat(paste(100*(1-alpha),"% Confidence Interval:", "(", lb, ",", ub, ")", sep=" "), "\n")
   cat("\n")
  cat(paste("One sided Test:H1.1 (1=reject H0, 0=cannot reject H0):", test1), "\n")
  cat(paste("Two sided Test:H1.2 (1=reject H0, 0=cannot reject H0):", test2), "\n")
  cat("\n\n\n")
  return(output)
  

}
